# -*- coding: utf-8 -*-
import pkg_resources
import pandas as pd 
import numpy as np
import math
import os
import re

from .process import load_word_pattern
 
"""
    texturizer.rhetoric: Feature flags indicating rhetorical devices

"""

########################################################################################

cliches_pat = load_word_pattern('cliches.dat')

jargon_pat = load_word_pattern('jargon.dat')

authority_pat = load_word_pattern('authority.dat')

########################################################################################
def add_text_rhetoric_features(df, columns):
    """
        Given a pandas dataframe and a set of column names.
        calculate the rhetoric trait features and add them.
    """
    rez = df.copy()
    for col in columns:
        rez = add_rhetoric_counts(rez, col)
    return rez

########################################################################################
def add_rhetoric_counts(df, col):
    """
        Given a pandas dataframe and a column name.
        Count the number of pattern matches for feature
    """
    df[col+'_cliches']=df[col].str.count(cliches_pat, flags=re.IGNORECASE)
    df[col+'_jargon']=df[col].str.count(jargon_pat, flags=re.IGNORECASE)
    df[col+'_authority']=df[col].str.count(authority_pat, flags=re.IGNORECASE)
    return df





# -*- coding: utf-8 -*-
import pandas as pd 
import numpy as np
import codecs
import re

from .process import load_word_list
from .process import load_word_pattern
from .process import remove_urls_and_tags
from .process import remove_escapes_and_non_printable

smiles = load_word_list("emoticons-smile.dat")
laughs = load_word_list("emoticons-laugh.dat")
winks = load_word_list("emoticons-wink.dat")
cheekys = load_word_list("emoticons-wink.dat")
kisses = load_word_list("emoticons-kiss.dat")
happycrys = load_word_list("emoticons-happy-cry.dat")
crys = load_word_list("emoticons-cry.dat")
sads = load_word_list("emoticons-sad.dat")
shocks = load_word_list("emoticons-shock.dat")
sceptics = load_word_list("emoticons-sceptical.dat")
 
fwd_regex = "[:;8BX]['’`]{0,1}[-=^oc]{0,2}[DPO0J3ox,Þþb@*\\|/()<>{}\[\]]{1,2}"
fwd_re = re.compile(fwd_regex)
bck_regex = "[@*\\|/()<>{}\[\]]{1,2}[-=^]{0,2}['’`]{0,1}[:;]"
bck_re = re.compile(bck_regex)

 
"""
    texturizer.emoticons: Emoticon Recognition Text Features
 
    The functions in this library will add columns to a dataframe that indivate
    whether there are emoticons in certain columns of text, and whether those
    emoticons represent one of the more common emotions.

    NOTE: In developing these regexes I have deliberately ignored certain emoticons 
     because of the likelihood of false positive matches in text containing brackets
     For example emoticons: 8) or (B will not be matched.

    To avoid matching characters inside document markup language tags there is a 
    rudimentary regex based tag removal and unescaped version of the text that is
    expecte to have been generated in the intial simple text function run by the program. 
    This will remove URLS and HTML tags before trying to match emoticons.

    Some references used when considering which empticons to include:

    https://www.unglobalpulse.org/2014/10/emoticon-use-in-arabic-spanish-and-english-tweets/

    https://www.researchgate.net/publication/266269913_From_Emoticon_to_Universal_Symbolic_Signs_Can_Written_Language_Survive_in_Cyberspace

    https://www.sciencedirect.com/science/article/abs/pii/S0950329317300939

    https://www.semanticscholar.org/paper/An-Approach-towards-Text-to-Emoticon-Conversion-and-Jha/3b81505fa7fec81563b2dafae3939fa1b07f3a98

    https://www.qualitative-research.net/index.php/fqs/article/view/175/391

    https://www.researchgate.net/publication/221622114_M_Textual_Affect_Sensing_for_Sociable_and_Expressive_Online_Communication

"""


########################################################################################
def add_text_emoticon_features(df, columns):
    """
        Given a pandas dataframe and a set of column names.
        Add features that detect the presence of emoticons.
    """
    rez = df.copy()
    for col in columns:
        rez = add_emoticon_features(rez, col)
    return rez


########################################################################################
def add_emoticon_features(df, col):
    """
        Given a pandas dataframe and a column name.
        Check for emoticons in the column and add a set of features
        that indicate both the presence and emotional flavour of the emoticon.
    """
    def cal_features(x, col):
        emos = 0 
        smiley = 0 
        wink = 0 
        kiss = 0
        happycry = 0
        laugh = 0
        cheeky = 0
        crying = 0
        sad = 0
        shock = 0 
        sceptic = 0 
        if x[col]==x[col]:
            text = remove_urls_and_tags( remove_escapes_and_non_printable( x[col] ) )
            matches = fwd_re.findall(text)
            bck_matches = bck_re.findall(text)
            if len(matches)>0 or len(bck_matches)>0:
                matches.extend(bck_matches)
                emos = len(matches)
                if set(matches).intersection( smiles ):                
                    smiley = 1
                if set(matches).intersection( crys ):        
                    crying = 1
                if set(matches).intersection( winks ):                
                    wink = 1
                if set(matches).intersection( kisses ):                
                    kiss = 1
                if set(matches).intersection( sads ):                
                    sad = 1
                if set(matches).intersection( shocks ):                
                    shock = 1
                if set(matches).intersection( sceptics ):                
                    sceptic = 1
                if set(matches).intersection( laughs ):                
                    laugh = 1
                if set(matches).intersection( cheekys ):                
                    cheeky = 1
                if set(matches).intersection( happycrys ):                
                    happycry = 1
        pos = smiley + wink + kiss + happycry + laugh + cheeky
        neg = crying + sad + shock + sceptic
        sent = pos - neg
        return emos,smiley,wink,kiss,happycry,laugh,cheeky,crying,sad,shock,sceptic,pos,neg,sent
 
    df[ get_emoticon_col_list(col) ] = df.apply(cal_features, col=col, axis=1, result_type="expand")

    return df

########################################################################################
def get_emoticon_col_list(col):
    return [col+'_emoticons', col+'_emo_smiley', col+'_emo_wink', col+'_emo_kiss', col+'_emo_happycry', col+'_emo_laugh', col+'_emo_cheeky', col+'_emo_cry', col+'_emo_sad', col+'_emo_shock', col+'_emo_sceptic', col+'_emo_pos', col+'_emo_neg', col+'_emo_sentiment'] 





# -*- coding: utf-8 -*-
import pkg_resources
import pandas as pd 
import numpy as np
import math
import os
import re

from .process import load_word_list
from .process import load_word_pattern

"""
    texturizer.literacy: Literacy feature flags

    Simple word matching to generate features for common literacy problems.
    This includes typos or spelling mistakes and some simple grammar problems,
    for example, not capitalizing the first word of a sentence.

"""

########################################################################################
misspelling_pat = load_word_pattern('misspelling.dat')
misspelling_re = re.compile(misspelling_pat)
 
grammar_pat = "[ '(\"][aA] [AaEeIiOoUu]|[^.][^A-Z]\. [a-z]|\b(\w+)\b \b\1\b"
grammar_re = re.compile(grammar_pat)

########################################################################################
def add_text_literacy_features(df, columns):
    """
        Given a pandas dataframe and a set of column names.
        calculate the simple literacy features and add them.
    """
    rez = df.copy()
    for col in columns:
        rez = add_literacy_features(rez, col)
    return rez

########################################################################################
def add_literacy_features(df, col):
    """
        Given a pandas dataframe and a column name.
        add simple text match features for literacy.
    """

    def lit_features(x, col):
        misspelling = 0
        grammar_err = 0
        if x[col]!=x[col]:
            misspelling = 0
        else:
            text = (x[col])
            grammar_err = len(grammar_re.findall(text))
            text = (x[col].lower())
            misspelling = len(misspelling_re.findall(text))
        return misspelling, grammar_err

    df[[ col+'_misspelling', col+'_grammar_err' ]] = df.apply(lit_features, col=col, axis=1, result_type="expand")

    return df
 



# -*- coding: utf-8 -*-
import pkg_resources
import pandas as pd 
import numpy as np
import math
import os
import re

from .process import load_word_pattern
 
"""
    texturizer.traits: Personality trait feature flags

    This module performs word or phrase matching to generate features 
    that can be indicative of personailuty traits in a writer or speaker.

    Some ideas taken from these articles

    https://www.scientificamerican.com/article/you-are-what-you-say/

    https://hbr.org/2011/12/your-use-of-pronouns-reveals-your-personality
"""

########################################################################################

reasoning_pat = load_word_pattern('reasoning.dat')
reasoning_re = re.compile(reasoning_pat)

nuance_pat = load_word_pattern('nuance.dat')
nuance_re = re.compile(nuance_pat)

explain_pat = load_word_pattern('explain.dat')
explain_re = re.compile(explain_pat)

singular_pat = "\\bi\\b|\\bme\\b|\\bmyself\\b|\\bmy\\b|\\bmine\\b"
singular_re = re.compile(singular_pat)

plural_pat = "\\bwe\\b|\\bus\\b|\\bour\\b|\\bourselves\\b"
plural_re = re.compile(plural_pat)

quotation_pat = "\"[ a-zA-Z0-9.,?!:;']*\""

########################################################################################
def add_text_trait_features(df, columns):
    """
        Given a pandas dataframe and a set of column names.
        calculate the personality trait features and add them.
    """
    rez = df.copy()
    for col in columns:
        rez = add_trait_counts(rez, col)
    return rez

########################################################################################
def add_trait_counts(df, col):
    """
        Given a pandas dataframe and a column name.
        Count the number of keyword matches for each trait
    """
    df[col+'_reason']=df[col].str.count(reasoning_pat, flags=re.IGNORECASE)
    df[col+'_explain']=df[col].str.count(explain_pat, flags=re.IGNORECASE)
    df[col+'_nuance']=df[col].str.count(nuance_pat, flags=re.IGNORECASE)
    df[col+'_singular']=df[col].str.count(singular_pat, flags=re.IGNORECASE)
    df[col+'_plural']=df[col].str.count(plural_pat, flags=re.IGNORECASE)
    df[col+'_quotations']=df[col].str.count(quotation_pat, flags=re.IGNORECASE)
    return df




# -*- coding: utf-8 -*-
import functools
import pandas as pd 
import numpy as np

from .process import start_profile
from .process import end_profile
from .simple import add_text_summary_features
from .pos import add_text_pos_features
from .topics import add_text_topics_features
from .profanity import add_text_profanity_features
from .traits import add_text_trait_features
from .rhetoric import add_text_rhetoric_features
from .sentiment import add_text_sentiment_features
from .literacy import add_text_literacy_features
from .emoticons import add_text_emoticon_features
from .comparison import add_comparison_features
from .scarcity import add_scarcity_features

"""
    texturizer.featurize: Core functions to apply a set of features to a data frame.
"""
########################################################################################

def process_df(df, params):
    """ 
    process_df: Function that co-ordinates the process of generating the features
    
    """ 
    start_profile("simple")
    simple = add_text_summary_features( df, params["columns"] )
    end_profile("simple")
    if params["comparison"] :
        start_profile("comparison")
        simple = add_comparison_features( simple, params["columns"] )
        end_profile("comparison")
    if params["profanity"] :
        start_profile("profanity")
        simple = add_text_profanity_features( simple, params["columns"] )
        end_profile("profanity")
    if params["sentiment"] :
        start_profile("sentiment")
        simple = add_text_sentiment_features( simple, params["columns"] )
        end_profile("sentiment")
    if params["scarcity"] :
        start_profile("scarcity")
        simple = add_scarcity_features( simple, params["columns"] )
        end_profile("scarcity")
    if params["emoticons"] :
        start_profile("emoticons")
        simple = add_text_emoticon_features( simple, params["columns"] )
        end_profile("emoticons")
    if params["topics"] :
        start_profile("topics")
        if params["count_matches"] :
            simple = add_text_topics_features( simple, params["columns"], 'count' )
        else:
            simple = add_text_topics_features( simple, params["columns"] )
        end_profile("topics")
    if params["traits"] :
        start_profile("traits")
        simple = add_text_trait_features( simple, params["columns"] )
        end_profile("traits")
    if params["rhetoric"] :
        start_profile("rhetoric")
        simple = add_text_rhetoric_features( simple, params["columns"] )
        end_profile("rhetoric")
    if params["pos"] :
        start_profile("pos")
        simple = add_text_pos_features( simple, params["columns"] )
        end_profile("pos")
    if params["literacy"] :
        start_profile("literacy")
        simple = add_text_literacy_features( simple, params["columns"] )
        end_profile("literacy")
    return simple

########################################################################################

def generate_feature_function(parameters):
    """
        This function will take the processed command line arguments that determine
        the feature to apply and partially apply them to the process_df function.
        Returning a function that can be used to apply those parameters to multiple
        chunks of a dataframe.
    """
    return functools.partial(process_df, params = parameters)




# -*- coding: utf-8 -*-
import pandas as pd 
import numpy as np
import math
import os
import re

from .process import remove_escapes_and_non_printable
from .process import remove_urls_and_tags
from .process import load_word_list

"""
    texturizer.simple: Basic text feature calculation.

    Calculate statistics such as the average length of words, max word length
    proportion of non stop-words. We also create a clean version of the text
    that can be used by other functions in the library.

    Stop-word list taken from: https://www.textfixer.com/tutorials/common-english-words.txt 

"""

stop_word_list = load_word_list("stop-words.dat") 

########################################################################################
def add_text_summary_features(df, columns):
    """
        Given a pandas dataframe and a set of column names.
        calculate the simple text summary features and add them.
    """
    rez = df.copy()
    for col in columns:
        rez = add_text_features(rez, col)
    return rez

########################################################################################
def add_text_features(df, col):
    """
        Given a pandas dataframe and a column name.
        calculate the simple text summary features and add them.
    """
    col_len = col + "_len"
    df[col_len] = df[col].apply(null_tolerant_len)
    def cal_features(x, col):
        if x[col]!=x[col]:
            word_count = 0 
            sentence_count = 0 
            line_count = 0 
            avg_word_len = 0 
            max_word_len = 0
            avg_sentence_len = 0 
            content_wd = 0 
            capital_d = 0
            punct_d = 0
            text = ""
        else:
            text = remove_urls_and_tags( remove_escapes_and_non_printable( x[col] ) )
            chars = null_tolerant_len(x[col])
            capitals = sum(1 for c in x[col] if c.isupper())
            punct = sum(1 for c in x[col] if c in ['.','!','?',':',';','-',','])
            capital_d = capitals/chars
            punct_d = punct/chars
            word_array = x[col].lower().split()
            sentence_array = [ x for x in re.split(r"[.?]", x[col].lower()) if x]
            line_array = [ x for x in re.split(r"[\r\n]+", x[col].lower()) if x]
            non_stop_words = list(set(word_array) - set(stop_word_list))
            word_count = len(word_array)
            sentence_count = len(sentence_array)
            line_count = len(line_array)
            word_lengths = list(map(len, word_array))
            max_word_len = max(word_lengths)
            avg_word_len = sum(word_lengths)/word_count
            content_wd = len(non_stop_words)/len(word_array)
        return word_count, sentence_count, line_count, avg_word_len, max_word_len, content_wd, capital_d, punct_d

    df[ get_simple_col_list(col) ] = df.apply(cal_features, col=col, axis=1, result_type="expand")

    return df

########################################################################################
def get_simple_col_list(col):
    return [col+'_wc', col+'_sc', col+'_lc', col+'_avg_wl', col+'_max_wl', col+'_cwd', col+'_caps', col+'_punc']

########################################################################################
def null_tolerant_len(x):
    if x != x:
        return 0
    else:
        return len(x)

 
########################################################################################



# -*- coding: utf-8 -*-
import pandas as pd 
import numpy as np
import math
import os
import re

"""
    texturizer.profanity: Profanity feature flags

    Simple word matching to generate features for common profanities, insults and slurs.

    I have generalised towards words that are unambiguously intended to insult or offend.
    There will therefore be some forms that are not cpatured because they have alternative
    uses in general English. For example: "whore", "spick", or "wog" are all words that 
    have legitimate applications where offence is not the intended purpose of usage.

    NOTE: That these have been deliberately reduced to a set that are more likely to be 
          encountered in modern Australian spoken language or internet comments.

"""

########################################################################################
 
pattern_start = "[ '(\"]fuck[^ ]*|[ '(\"]cunt[^ ]*[ .,?;:']|[- '(\"]"
hard_profanity_list = ["cockhead","bastard","bitch","shit","piss","piss-off","pissing","piss-poor","cocksucker", "cocksucka", "shits","shitloads", "shitfaced", "shithead", "shitlist"]
hard_pat = pattern_start  + ( "[ .,?!;:'\"]|[- '(\"]".join(hard_profanity_list) ) + "[ .,?;:']"
hard_re = re.compile(hard_pat)

mild_profanity_list = ["bugger","bloody","crap","damnit","sod","goddamn","wanker","bollocks","bullshit"]

non_profanity_list = ["fiddlesticks","blimey" ,"zounds","gad","gadzooks","gosh","golly","darn","tarnation","dang","jiminy","drat","crikey","doggone","heck","gee","gorblimey","jeez","jeepers","jeepers-creepers"]

insult_list = ["twat", "moron", "dweeb", "cretin", "idiot", "dickhead", "cockhead", "cocksucker", "cocksucka", "cuck", "faggot", "poofter", "poofta", "cuntboy", "fuckwit", "arsehole", "bastard", "bitch", "douchebag", "douchnozzle", "shitgibbon", "cockwaffle", "shithead", "twatwaffle", "fucktrumpet", "pisswizard","fuckface", "fuckhead", "sissy", "slut", "soyboy", "soy-boy", "shitfaced", "shithead", "shitbag", "shitstain", "shithole","dumbass","motherfucker", "dumbfuck", "fuckknuckle", "fucknuckle", "motherfucka", "muthafucka", "redneck", "red-neck", "whitetrash", "white-trash", "trailertrash", "trailer-trash"]

slur_list = ["abo","abbo","boong", "chinaman", "chonky", "curry-muncher", "darky","darkey","darkie", "gook", "guizi", "gweilo", "half-breed", "nigger", "raghead", "rag-head", "sandnigger", "sand-nigger", "sambo", "slanteye", "slant-eye", "slopehead", "slope-head","towelhead", "towel-head","spearchucker","spear-chucker", "wigger", "whigger", "wigga", "wop"] 

masked_pat = "\\bf[*%$#@c?]{2}k|\\bs[*%$#@h?1!4]{2}t|\\bc[*%$#@n]{2}t|\\b@$$|\\b$h1t|\\bb!tch|\\bbi\+ch|\bc0ck|\\bl3itch\\b|\\bp\*ssy\\b|\\bdik\\b"
masked_re = re.compile(masked_pat)

########################################################################################
def add_text_profanity_features(df, columns):
    """
        Given a pandas dataframe and a set of column names.
        calculate the simple text summary features and add them.
    """
    rez = df.copy()
    for col in columns:
        rez = add_profanity_features(rez, col)
    return rez

########################################################################################
def add_profanity_features(df, col):
    """
        Given a pandas dataframe and a column name.
        add simple text match features for profanities.
    """

    def prof_features(x, col):
        hard_profanity = 0
        mask_profanity = 0
        contains_insult = 0
        mild_profanity = 0
        non_profanity = 0
        contains_slur = 0
        if x[col]!=x[col]:
            mild_profanity = 0
        else:
            text = (x[col].lower())
            word_array = text.split()
            hard_profanity = len(hard_re.findall(text))
            mask_profanity = len(masked_re.findall(text))
            if set(mild_profanity_list).intersection(word_array):
                mild_profanity = 1
            if set(non_profanity_list).intersection(word_array):
                non_profanity = 1
            if set(insult_list).intersection(word_array):
                contains_insult = 1
            if set(slur_list).intersection(word_array):
                contains_slur = 1
        return hard_profanity, mask_profanity, mild_profanity, non_profanity, contains_insult, contains_slur

    df[ get_profanity_col_list(col) ] = df.apply(prof_features, col=col, axis=1, result_type="expand")

    return df
 

########################################################################################
def get_profanity_col_list(col):
    return [ col+'_profane_hard', col+'_profane_masked', col+'_profane_mild', col+'_profane_non', col+'_insult', col+'_slur' ]




import sys
import numpy as np
import jellyfish
import textdistance
import pandas as pd

#################################################################################
def add_comparison_features(df,columns):
    """
        This is the entry point to add all the core text similarity features.
        Note: We left out Ratcliff Obershelp from the set of metrics because it
        takes close to an order of magnitude longer to compute.

        Initial version just includes 4 string edit distance metrics. 
    """
    return add_string_match_features(df,columns)

#################################################################################
def add_string_match_features(df,columns):
    """
    Return a copy of a dataframe with features describing matching
    between the set of named text columns
    """
    def sm_features(x, col1, col2):
        if (x[col1] != x[col1]) or (x[col2] != x[col2]):
            jd = np.nan
            ld = np.nan
            ji = np.nan
            sd = np.nan
        else:
            raw_text1 = x[col1].lower()
            raw_text2 = x[col2].lower()
            jd = jellyfish.jaro_distance(raw_text1,raw_text2)
            ld = jellyfish.levenshtein_distance(raw_text1,raw_text2)
            ji = textdistance.jaccard(raw_text1,raw_text2)
            sd = textdistance.sorensen(raw_text1,raw_text2 )
        return jd, ld, ji, sd

    col_number = len(columns)
    for i in range( col_number-1 ):
        for j in range(i+1,col_number):
            col1 = columns[i]
            col2 = columns[j]
            prefix = col1 + "_vs_" + col2
            df[[prefix+'_jd', prefix+'_ld',prefix+'_ji',prefix+'_sd']] = df.apply(sm_features, col1=col1,col2=col2, axis=1, result_type="expand")
    return df


#################################################################################
def add_ratcliff_obershelp(df,columns):
    """
    Return a copy of a dataframe with features describing matching
    between the set of named text columns
    """
    def sm_features(x, col1, col2):
        if (x[col1] != x[col1]) or (x[col2] != x[col2]):
            ro = np.nan
        else:
            raw_text1 = x[col1].lower()
            raw_text2 = x[col2].lower()
            ro = textdistance.ratcliff_obershelp(raw_text1,raw_text2)
        return ro

    col_number = len(columns)
    for i in range( col_number-1 ):
        for j in range(i+1,col_number):
            col1 = columns[i]
            col2 = columns[j]
            prefix = col1 + "_vs_" + col2
            df[[prefix+'_rat_obers']] = df.apply(sm_features, col1=col1,col2=col2, axis=1, result_type="expand")
    return df






from sklearn.base import TransformerMixin, BaseEstimator
import pandas as pd
import numpy as np
import re

from .featurize import generate_feature_function

class TextTransform(TransformerMixin, BaseEstimator):
    """
        This class implements a SciKit Learn compatible Transformer for
         convert a text column into a series of numeric values.
         You specify the transformations you want as an array of named
         feature groups.

        columns: Array(String) Names of the text columns to process.
        transforms: Array(String) Names of the feature sets to generate
                     Options: profanity, topics, pos, rhetoric, literacy, traits
    """

    def __init__(self, columns, transforms=['simple']):
        self.transforms = transforms
        self.columns = columns
        self.config = self.generate_feature_config(columns, transforms)
        self.func = generate_feature_function(self.config)
        

    def fit(self, X, y=None, **fit_params):
        return self
    
    def transform(self, X, y=None, **transform_params):
        """
            Transform the matrix of values
             -- need to deal with single or multiple columns
        """
        rez = self.func(X)        

        # Might need this later 
        #if X.__class__.__name__ == "DataFrame":
        #    X = X.values

        # REMOVE THE TEXT COLUMNS -- PARAMETERIZE THIS LATER
        for col in self.columns:
            rez.drop(col, inplace=True, axis=1)

        return rez 


    #############################################################
    def generate_feature_config(self, columns, params):
        """
        We need to process the params into a particular data structure 
        for the dataframe processing function to recognize. 
        """
        result = {
              "columns":columns,
              "profanity":False,
              "sentiment":False,
              "emoticons":False,
              "topics":False,
              "count_matches":False,
              "traits":False,
              "rhetoric":False,
              "pos":False,
              "literacy":False,
              "scarcity":False,
              "comparison":False,
              "embedding":False
        }
        if "profanity" in params:
           result["profanity"]=True
        if "sentiment" in params:
           result["sentiment"]=True
        if "scarcity" in params:
           result["scarcity"]=True
        if "emoticons" in params:
           result["emoticons"]=True
        if "topics" in params:
           result["topics"]=True
        if "count_matches" in params:
           result["count_matches"]=True
        if "pos" in params:
           result["pos"]=True
        if "traits" in params:
           result["traits"]=True
        if "rhetoric" in params:
           result["rhetoric"]=True
        if "literacy" in params:
           result["literacy"]=True
        if "comparison" in params:
           result["comparison"]=True
        if "embedding" in params:
           result["embedding"]=True

        return result





# -*- coding: utf-8 -*-
 
""" texturizer.texturizer: provides entry point main()."""
 
import pandas as pd
import sys
import os

from .process import load_complete_dataframe
from .process import process_file_in_chunks
from .process import initialise_profile
from .process import print_profiles
from .process import print_output

from .featurize import generate_feature_function

from .config import max_filesize
 
def main():
    """Main texturizer application entry point.
       parses out CL options and determine the size of the file.
       Then process the file for the requested features
    """
    if len(sys.argv) < 2:
        print("ERROR: MISSING ARGUMENTS")
        print_usage(sys.argv)
        exit(1)
    else:
        params = get_cmd_line_params(sys.argv)

        if not os.path.exists(params["dataset"]):
            print("ERROR: Dataset does not exist")
            print_usage(sys.argv)
            exit(1)

        initialise_profile()
        feature_func = generate_feature_function(params)

        filesize = os.stat(params["dataset"]).st_size
        if filesize<max_filesize:
            df = load_complete_dataframe( params["dataset"] )
            simple = feature_func(df)
            print_output( simple )
        else:
            process_file_in_chunks(params["dataset"], feature_func)

        print_profiles()


#############################################################
def get_cmd_line_params(argv):
    """ parse out the option from an array of command line arguments """
    data = argv[-1]
    options = argv[1:-1]
    result = {"dataset":data,
              "columns":[], 
              "profanity":False, 
              "sentiment":False, 
              "emoticons":False, 
              "topics":False, 
              "count_matches":False, 
              "traits":False, 
              "rhetoric":False, 
              "pos":False, 
              "literacy":False, 
              "scarcity":False, 
              "comparison":False, 
              "embedding":False
    }
    for o in options:
        parts = o.split("=")
        if parts[0] == "-literacy":
            result["literacy"]=True
        if parts[0] == "-profanity":
            result["profanity"]=True
        if parts[0] == "-scarcity":
            result["scarcity"]=True
        if parts[0] == "-sentiment":
            result["sentiment"]=True
        if parts[0] == "-topics":
            result["topics"]=True
            if len(parts)>1:
                if parts[1] == 'count':
                    result["count_matches"]=True
        if parts[0] == "-traits":
            result["traits"]=True
        if parts[0] == "-rhetoric":
            result["rhetoric"]=True
        if parts[0] == "-pos":
            result["pos"]=True
        if parts[0] == "-emoticons":
            result["emoticons"]=True
        if parts[0] == "-embedding":
            result["embedding"]=True
        if parts[0] == "-comparison":
            result["comparison"]=True
        if parts[0] == "-columns":
            cols = parts[1].split(",")
            result["columns"]=cols

    return result

#############################################################
def print_usage(args):
    """ Command line application usage instrutions. """
    print("USAGE ")
    print(args[0], " [ARGS] <PATH TO DATASET>")
    print("  <PATH TO DATASET> - Supported file types: csv, tsv, xls, xlsx, odf")
    print(" [ARGS] In most cases these are switches that turn on the feature type")
    print("  -columns=<COMMA SEPARATED LIST>. REQUIRED")
    print("  -topics OR -topics=count. Default: False. Match words from common topics (or count matches).")
    print("  -traits Default: False. Word usage for personality traits.")
    print("  -rhetoric Default: False. Word usage for rhetorical devices.")
    print("  -pos Default: False. Part of speech proportions.")
    print("  -literacy Default: False. Checks for common literacy markers.")
    print("  -profanity Default: False. Profanity check flags.")
    print("  -sentiment Default: False. Words counts for positive and negative sentiment.")
    print("  -scarcity Default: False. Word scarcity scores.")
    print("  -emoticons Default: False. Emoticon check flags.")
    print("  -comparison Default: False. Cross-column comparisons.")
    print("")





# -*- coding: utf-8 -*-

import spacy
from .process import eprint

try:
   nlp = spacy.load("en_core_web_sm")
except:
   eprint(" * WARNING: POS features require the SpaCY language model : en_core_web_sm")

"""
    texturizer.pos Part of Speech features using spaCy

    Extraction of part of speech tags for a block of text and then generation
    of numerical features that summarise the grammatical structure of that text.

    Notes: I have implemented this inside a single 'apply' function which gives 
    some speed advantage. However, it is still slow. I am still looking for ways
    to speed this up.
"""

########################################################################################
def add_text_pos_features(df, columns):
    """
        Given a pandas dataframe and a set of column names.
        calculate the part of speech features and add them.
    """
    rez = df.copy()
    for col in columns:
        rez = add_pos_features(rez, col)
    return rez

########################################################################################
def add_pos_features(df, col):
    """
        Given a pandas dataframe and a column name.
        add features for the proportion of dominant parts of speech
        Nouns, Verbs, Adjectives, Adverbs, Pronouns and Adpositions
    """

    def pos_features(x, col):
        nouns = 0
        verbs = 0
        adj = 0
        adv = 0
        pron = 0
        adp = 0
        index = 1
        if x[col]!=x[col]:
            nouns = 0
        else:
            text = (x[col])
            doc = nlp(text)
            index = 0
            for token in doc:
               if token.pos_ == "VERB":
                   verbs = verbs + 1
               if token.pos_ == "NOUN":
                   nouns = nouns + 1
               if token.pos_ == "ADJ":
                   adj = adj + 1
               if token.pos_ == "ADV":
                   adv = adv + 1
               if token.pos_ == "PRON":
                   pron = pron + 1
               if token.pos_ == "ADP":
                   adp = adp + 1
               index = index + 1
        return nouns/index, verbs/index, adj/index, adv/index, pron/index, adp/index

    df[[ col+'_nouns', col+'_verbs', col+'_adj', col+'_adv', col+'_pron', col+'_adp' ]] = df.apply(pos_features, col=col, axis=1, result_type="expand")

    return df
 



# -*- coding: utf-8 -*-
import pkg_resources
import pandas as pd 
import numpy as np
import math
import os
import re

from .process import load_word_list
from .process import load_word_pattern

"""
    texturizer.topics: Common Topic Features

    Simple pattern matching to generate features for very common topics for text data.
    We have focused on topics that are common in both traditional and social media, 
    as well as commentary and discussion by the general public.
  
    The goal with these features is to provide a count of words which are 
    unambigously related to a specific topic.

    NOTE: That the words in these sets have been deliberately selected to be a set 
          that are less likely to match false positives. In other words they are 
          generally only used when talking about that specific topic, or when talking
          using metaphor or analogy.
"""

########################################################################################

religion_pat = load_word_pattern('religion.dat', "\\bchristi|\\bislam|")
religion_re = re.compile(religion_pat)

politics_pat = load_word_pattern('politics.dat', "\\bpoliti|")
politics_re = re.compile(politics_pat)

sex_pat = load_word_pattern('sex.dat', "\\bsex[^t]|")
sex_re = re.compile(sex_pat)

ethno_pat = load_word_pattern('ethnicity.dat', "\\bethn|")
ethno_re = re.compile(ethno_pat)

health_pat = load_word_pattern('health.dat', "\\bhealth|")
health_re = re.compile(health_pat)
 
econo_pat = load_word_pattern('economics.dat', "\\becono|\\bfinan|")
econo_re = re.compile(econo_pat) 
 
sport_pat = load_word_pattern('sports.dat', "\\bathlet|")
sport_re = re.compile(sport_pat)
 
arts_pat = load_word_pattern('arts.dat', "\\bartist|")
arts_re = re.compile(arts_pat)

family_pat = load_word_pattern('family.dat', "\\bfamil|")
family_re = re.compile(family_pat)
 
love_pat = load_word_pattern('love.dat', "\\bromanc|")
love_re = re.compile(love_pat)
 
crime_pat = load_word_pattern('crime.dat', "\\bcrimina|")
crime_re = re.compile(crime_pat)
 
travel_pat = load_word_pattern('travel.dat', "\\btravel|")
travel_re = re.compile(travel_pat)
 
food_pat = load_word_pattern('food.dat')
food_re = re.compile(food_pat)
 
technology_pat = load_word_pattern('technology.dat', "\\btechnol[^t]|")

fashion_pat = load_word_pattern('fashion.dat', "\\bfashion[^i]|")

culture_pat = load_word_pattern('culture.dat', "\\bcultur|")

education_pat = load_word_pattern('education.dat', "\\beducat|")

science_pat = load_word_pattern('science.dat', "\\bscientifi|")

########################################################################################
def add_text_topics_features(df, columns, type="flag"):
    """
        Given a pandas dataframe and a set of column names.
        calculate the simple text summary features and add them.
    """
    rez = df.copy()
    for col in columns:
        if type=="count":
            rez = add_topic_counts(rez, col)
        else:
            rez = add_topic_indicators(rez, col)
    return rez

########################################################################################
def add_topic_indicators(df, col):
    """
        Given a pandas dataframe and a column name.
        add simple text match for top indicators.
    """ 
    df[ col+'_religion' ] = 0
    df.loc[(df[col].notnull()) & (df[col].str.contains(religion_pat)), col+'_religion' ] = 1
    df[ col+'_politics' ] = 0
    df.loc[(df[col].notnull()) & (df[col].str.contains(politics_pat)), col+'_politics' ] = 1
    df[ col+'_sex' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(sex_pat)), col+'_sex' ]=1
    df[ col+'_ethnicity' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(ethno_pat)), col+'_ethnicity' ]=1
    df[ col+'_economics' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(econo_pat)), col+'_economics' ]=1
    df[ col+'_health' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(health_pat)), col+'_health' ]=1
    df[ col+'_sport' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(sport_pat)), col+'_sport' ]=1
    df[ col+'_arts' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(arts_pat)), col+'_arts' ]=1
    df[ col+'_family' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(family_pat)), col+'_family' ]=1
    df[ col+'_love' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(love_pat)), col+'_love' ]=1
    df[ col+'_crime' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(crime_pat)), col+'_crime' ]=1
    df[ col+'_travel' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(travel_pat)), col+'_travel' ]=1
    df[ col+'_food' ]=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(food_pat)), col+'_food' ]=1
    df[col+'_technology']= 0
    df.loc[(df[col].notnull()) & (df[col].str.contains(technology_pat)), col+'_technology' ]=1
    df[col+'_fashion']=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(fashion_pat)), col+'_fashion' ]=1
    df[col+'_culture']=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(culture_pat)), col+'_culture' ]=1
    df[col+'_education']=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(education_pat)), col+'_education' ]=1
    df[col+'_science']=0
    df.loc[(df[col].notnull()) & (df[col].str.contains(science_pat)), col+'_science' ]=1

    return df

########################################################################################
def add_topic_counts(df, col):
    """
        Given a pandas dataframe and a column name.
        Count the number of keywor matches for each topic
    """
    df[col+'_religion']=df[col].str.count(religion_pat, flags=re.IGNORECASE)
    df[col+'_politics']=df[col].str.count(politics_pat, flags=re.IGNORECASE)
    df[col+'_sex']=df[col].str.count(sex_pat, flags=re.IGNORECASE)
    df[col+'_ethnicity']=df[col].str.count(ethno_pat, flags=re.IGNORECASE)
    df[col+'_economics']=df[col].str.count(econo_pat, flags=re.IGNORECASE)
    df[col+'_health']=df[col].str.count(health_pat, flags=re.IGNORECASE)
    df[col+'_sport']=df[col].str.count(sport_pat, flags=re.IGNORECASE)
    df[col+'_arts']=df[col].str.count(arts_pat, flags=re.IGNORECASE)
    df[col+'_family']=df[col].str.count(family_pat, flags=re.IGNORECASE)
    df[col+'_love']=df[col].str.count(love_pat, flags=re.IGNORECASE)
    df[col+'_crime']=df[col].str.count(crime_pat, flags=re.IGNORECASE)
    df[col+'_travel']=df[col].str.count(travel_pat, flags=re.IGNORECASE)
    df[col+'_food']=df[col].str.count(food_pat, flags=re.IGNORECASE)
    df[col+'_technology']=df[col].str.count(technology_pat, flags=re.IGNORECASE)
    df[col+'_fashion']=df[col].str.count(fashion_pat, flags=re.IGNORECASE)
    df[col+'_culture']=df[col].str.count(culture_pat, flags=re.IGNORECASE)
    df[col+'_education']=df[col].str.count(education_pat, flags=re.IGNORECASE)
    df[col+'_science']=df[col].str.count(science_pat, flags=re.IGNORECASE)
    return df




# -*- coding: utf-8 -*-
from textblob import TextBlob
import numpy as np 
import re

from .process import load_word_pattern

"""
    texturizer.sentiment: Sentiment feature flags

    We use multiple sets of words lists for different sentiment features


    Hu and Liu

    This is a large list of postive and negative words. Taken from:
    http://www.cs.uic.edu/~liub/FBS/sentiment-analysis.html

    Minqing Hu and Bing Liu. "Mining and Summarizing Customer Reviews."
       Proceedings of the ACM SIGKDD International Conference on Knowledge
       Discovery and Data Mining (KDD-2004), Aug 22-25, 2004, Seattle,
       Washington, USA,

    AFINN

    Smaller list created by extracting extreme sentiment words from the AFINN
    lists. Extracted from:
    http://www2.imm.dtu.dk/pubdb/pubs/6010-full.html 

    This derived database of words is copyright protected and distributed under
    "Open Database License (ODbL) v1.0"
    http://www.opendatacommons.org/licenses/odbl/1.0/ 


    TODO

    Investgate how to use (license and processing requirements).

    http://www.wjh.harvard.edu/~inquirer/homecat.htm
"""

########################################################################################

positive_pat_large = load_word_pattern('positive-words.dat')
negative_pat_large = load_word_pattern('negative-words.dat')

positive_pat = load_word_pattern('positive.dat')
negative_pat = load_word_pattern('negative.dat')

########################################################################################
def add_text_sentiment_features(df, columns):
    """
        Given a pandas dataframe and a set of column names.
        calculate the sentiment features and add them.
    """
    rez = df.copy()
    for col in columns:
        rez = add_sentiment_features(rez, col)
        rez = add_textblob_features(rez, col)
    return rez

########################################################################################
def add_textblob_features(df, col):
    def tb_features(x, col):
        if x[col]!=x[col]:
            subjectivity = 0.0 #np.nan
            polarity = 0.0     #np.nan
        else:
            text = ( x[col] )
            blob = TextBlob(text)
            subjectivity = blob.sentiment.subjectivity
            polarity = blob.sentiment.polarity
        return polarity, subjectivity

    df[[ col+'_tb_polarity', col+'_tb_subjectivity' ]] = df.apply(tb_features, col=col, axis=1, result_type="expand")

    return df

########################################################################################
def add_sentiment_features(df, col):
    """
        Given a pandas dataframe and a column name.
        add simple text match features for sentiment.
    """
    wc_col = col+'_wc' # This is ALWAYS computed first
    df[col+'_positive']  = df[col].str.count(positive_pat, flags=re.IGNORECASE).fillna(0)
    df[col+'_negative']  = df[col].str.count(negative_pat, flags=re.IGNORECASE).fillna(0)
    df[col+'_sentiment'] = (df[col+'_positive'] - df[col+'_negative'] )/df[wc_col]

    return df 




# -*- coding: utf-8 -*-
 
"""setup.py: setuptools control."""
 
import re
from setuptools import setup
 
version = re.search(
        '^__version__\s*=\s*"(.*)"',
        open('texturizer/__init__.py').read(),
        re.M
    ).group(1)
 
with open("README.md", "rb") as f:
    long_descr = f.read().decode("utf-8")

setup(
      license="MIT",
      name = "texturizer",
      packages = ["texturizer"],
      install_requires=[
        'pandas','numpy','sklearn','jellyfish','textdistance','spacy','textblob'
      ],
      include_package_data=True,
      entry_points = {
        "console_scripts": ['texturizer = texturizer.texturizer:main']
      },
      version = version,
      description = "Python command line application to add text features to a CSV or TSV dataset.",
      long_description = long_descr,
      long_description_content_type='text/markdown',
      author = "John Hawkins",
      author_email = "johnc@getting-data-science-done.com",
      url = "http://john-hawkins.github.io",
      project_urls = {
          'Documentation': "http://texturizer.readthedocs.io",
          'Source': 'https://github.com/john-hawkins/texturizer',
          'Tracker': 'https://github.com/john-hawkins/texturizer/issues',
      },
    )





# -*- coding: utf-8 -*-
import pkg_resources
import pandas as pd 
import numpy as np
import statistics
import string
import math
import os
import re

from .process import load_dictionary
from .process import remove_escapes_and_non_printable
from .process import remove_urls_and_tags
 
"""
    texturizer.scarcity: Word Scarcity Statistics

    This module performs word scarcity analysis
"""

########################################################################################

scarcity = load_dictionary('scarcity.dat')

########################################################################################
def add_scarcity_features(df, columns):
    """
        Given a pandas dataframe and a set of column names.
        calculate the scarcity statistics features and add them.
    """
    rez = df.copy()
    for col in columns:
        rez = add_scarcity(rez, col)
    return rez

########################################################################################
def get_scarcity(word):
    if word in scarcity:
        return scarcity[word]
    else:
        return 1.0

########################################################################################
def add_scarcity(df, col):
    """
        Given a pandas dataframe and a column name.
    """
    def cal_features(x, col):
        if x[col]!=x[col]:
            mean_scarcity = 0
            median_scarcity = 0
            max_scarcity = 0
        else:
            text = remove_urls_and_tags( remove_escapes_and_non_printable( x[col] ) ).lower()
            text = text.translate(str.maketrans('', '', string.punctuation)).lower()
            words = text.split()
            scarcities = list(map(get_scarcity, words))
            mean_scarcity = statistics.mean(scarcities)
            median_scarcity = statistics.median(scarcities)
            max_scarcity = max(scarcities)
        return mean_scarcity, median_scarcity, max_scarcity 
    df[ get_scarcity_col_list(col) ] = df.apply(cal_features, col=col, axis=1, result_type="expand") 

    return df

########################################################################################
def get_scarcity_col_list(col):
    return [col+'_scarcity_mean', col+'_scarcity_median', col+'_scarcity_max']





# -*- coding: utf-8 -*-
from __future__ import print_function
from io import StringIO
import datetime as dt
import pkg_resources
import pandas as pd 
import numpy as np
import json
import codecs
import sys
import os
import re

from .config import max_filesize

"""
    texturizer.process: Support functions for the texturizer package.
                        Including functions for loading word lists, logging processing times,
                        and iteratively processing a large dataset in chunks.
"""
########################################################################################
resource_package = __name__

def load_dictionary(filename, escape=False):
    """
    Utility function to load a json serialised dictionary
    """
    _path = '/'.join(('data', filename))
    rawd = pkg_resources.resource_string(resource_package, _path).decode("utf-8")
    if escape:
        rawd = re.escape(rawd)
    return json.loads(str(rawd)) 

########################################################################################
def load_word_list(filename, escape=False):
    """
    Utility function to load topic vocab word lists for pattern matching.
    """
    _path = '/'.join(('data', filename))
    rawd = pkg_resources.resource_string(resource_package, _path).decode("utf-8")
    rawd = rawd[:-1]
    if escape:
        rawd = re.escape(rawd)
    word_list = str(rawd).split(r"\n")
    _list = [i for i in word_list if i]
    return _list

########################################################################################

def load_word_pattern(filename, prefix="", pluralize=True, bound=True, escape=False):
    word_list = load_word_list(filename, escape=escape)
    if bound:
       delimiter = "\\b"
    else:
       delimiter = ""
    pattern_start = prefix + delimiter
    if pluralize:
        tail = "s*" + delimiter
        joiner = "s*" + delimiter + "|" + delimiter
    else:
        tail = delimiter
        joiner = delimiter + "|" + delimiter

    pattern = pattern_start  + ( joiner.join(word_list) ) + tail
    return pattern

########################################################################################
"""
   This is a set of functions to allow the application to print time
   profiles of the various feature engines to STDERR.
"""

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

profiles = {}

def initialise_profile():
    profiles = {}

def start_profile(proc_name):
    n1=dt.datetime.now()
    if proc_name in profiles:
        profiles[proc_name]["start"] = n1
    else:
        profiles[proc_name] = {"start":n1}

def end_profile(proc_name):
    n2 = dt.datetime.now()
    n1 = profiles[proc_name]["start"]
    total = n2-n1
    profiles[proc_name]["end"] = n2
    if "total" in profiles[proc_name]:
        curr_total = profiles[proc_name]["total"]
        profiles[proc_name]["total"] = curr_total + total
    else:
        profiles[proc_name]["total"] = total

def print_profiles():
    eprint("Computation Time Profile for each Feature Set")
    eprint("---------------------------------------------")
    for k in profiles.keys():
        eprint(padded(k), str(profiles[k]["total"]) ) 

def padded(k, padto=20):
    spacer_len = padto - len(k)
    return k + (" "*spacer_len)

########################################################################################
def process_file_in_chunks(path_to_file, function_to_apply):
    """
        Given a path to a large dataset we will iteratively load it in chunks and 
        apply the supplied function to and write the result to the output stream.
    """
    fsize = os.stat(path_to_file).st_size
    sample_prop = max_filesize / fsize 
    line_count = count_lines(path_to_file)
    chunks = round(line_count * sample_prop)
    data_iterator = pd.read_csv(path_to_file, chunksize=chunks, low_memory=False)
    total_chunks = 0
    for index, chunk in enumerate(data_iterator, start=0):
        startpoint = 0 + (index*chunks)
        total_chunks = index + 1
        temp = function_to_apply(chunk)
        if total_chunks==1:
            print_output(temp, header=True)
        else:
            print_output(temp, header=False)
    
    eprint("Chunks processed: ", total_chunks)
 
########################################################################################
def print_output(df, header=True):
    output = StringIO()
    df.to_csv(output,index=False, header=header)
    print(output.getvalue(), end = '')

########################################################################################
def extract_file_extension(path_to_file):
    return os.path.splitext(path_to_file)[1]

########################################################################################
def load_complete_dataframe(path_to_file):
    """
        We load the entire dataset into memory, using the file extension to determine
        the expected format. We are using encoding='latin1' because it appears to 
        permit loading of the largest variety of files.
        Representation of strings may not be perfect, but is not important for generating a
        summarization of the entire dataset.
    """
    extension = extract_file_extension(path_to_file).lower()
    if extension == ".csv":
        df = pd.read_csv(path_to_file, encoding='latin1', low_memory=False)
        return df
    if extension == ".tsv":
        df = pd.read_csv(path_to_file, encoding='latin1', sep='\t', low_memory=False)
        return df
    if extension == ".xls" or extension == ".xlsx" or extension == ".odf" :
        df = pd.read_excel(path_to_file)
        return df

    raise ValueError("Unsupported File Type")

########################################################################################
def count_lines(path_to_file):
    """
    Return a count of total lines in a file. In a way that filesize is irrelevant
    """
    count = 0
    for line in open(path_to_file): count += 1
    return count

########################################################################################
def len_or_null(val):
    """ 
       Alternative len function that will simply return numpy.NA 
       for invalid values. This is needed to get sensible results 
       when running len over a column that may contain nulls
    """
    try:
        return len(val)
    except:
        return np.nan

########################################################################################
def isNaN(num):
    return num != num

########################################################################################
def remove_urls_and_tags(text):
    """
        Remove any obvious text elements that appear to be either 
        URLs or HTML tags
    """
    patterns = ["https?://[-._a-z0-9A-Z]*","</?[a-zA-Z]* ?[a-zA-Z'=.]* ?/?>","\\[tnrf]"]
    new_text = text
    for p in patterns:
        new_text = re.sub(p, ' ', new_text)
    return new_text

def remove_urls(text):
    pattern = "https?://[-._a-z0-9A-Z]*"
    new_text = re.sub(pattern, ' ', text)
    return new_text

def remove_tags(text):
    pattern = "</?[a-zA-Z]* ?[a-zA-Z'=.]* ?/?>"
    new_text = re.sub(pattern, ' ', text)
    return new_text

########################################################################################
def remove_escapes_and_non_printable(text):
    """
        Apply the codecs escape to decode any escaped characters.
        Then apply a regex to remove any non printable characters
    """
    try:
        new_text0 = codecs.escape_decode(text)[0].decode("utf-8")
    except:
        new_text0 = text
    pattern = "\0|\n|\r|\b|\t|\f|\v"
    new_text1 = re.sub(pattern, " ", new_text0)
    return new_text1
